#!/usr/bin/env python3

import os
import sys
import datetime
import re
import configparser
import json
import argparse
import pika
import ssl
from . import version as VERSION
from .mhclient import MhClient

# define globals
state = {}


class InvalidSearchKey(Exception):
    pass


class InvalidResourceName(Exception):
    pass


class InvalidDate(Exception):
    pass


class ResourceNotFound(Exception):
    pass


def formdate(dt):
    """Format datetime for printing to parsable output"""
    return dt.strftime("%Y%m%d%H%M")

def parseuserdate(date):
    """Parse user date format into a datetime object (todo raise custom
    error for wrong date)
    """
    if date is None: return None
    if date == "": return None
    try:
        datel = re.sub("[-.:T ]", "", date)
        datel = datel[:12]
        datel = datel+(12-len(datel))*"0"
        return datetime.datetime.strptime(datel, "%Y%m%d%H%M")
    except:
        raise InvalidDate(date)

def parsemhdate(date):
    """Parse various date formats encountered in meteohub into a datetime
    object
    """
    return datetime.datetime.strptime(date.strip("Z"),"%Y-%m-%dT%H:%M:%S.%f")

def matchdate(ud, md):
    """Match user and mh date"""
    if ud is None: return True
    return ud == md

def parse_search_key(search_key):
    """Parse schedule search key"""
    if search_key == "":
        raise InvalidSearchKey()
    name = None
    date = None
    kl = search_key.split(":")
    try:
        name = kl[0]
        date = kl[1]
    except (ValueError, IndexError):
        pass
    srft = parseuserdate(date)
    return name, srft


def amqp_connect(username, password, host, port, ssl_enabled):
    credentials = pika.PlainCredentials(username, password)
    if ssl_enabled:
        context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
        context.verify_mode = ssl.CERT_NONE
        context.load_default_certs()
        ssl_options = pika.SSLOptions(
            context=context, server_hostname=host
        )
    else:
        ssl_options = None

    # get connection
    if state["verbose"]:
        print(f"Connecting to the AMQP Server {host}:{port} (SSL = {ssl_enabled})")
    try:
        connection = pika.BlockingConnection(
            pika.ConnectionParameters(
                host=host,
                port=int(port),
                virtual_host='/',
                credentials=credentials,
                ssl_options=ssl_options))
        channel = connection.channel()
        return channel, connection
    except pika.exceptions.AMQPConnectionError as e:
        exit(f"Failed to send data, connection failed: {e}")
    except pika.exceptions.AMQPChannelError as e:
        exit(f"Failed to send data, channel is dead: ({e})")


def settings(args): # section_name: str):
    """Print app configuration"""
    if args.section_name in state["config"]:
        for key in state["config"][args.section_name]:
            print(f"{key} = {state['config'][args.section_name][key]}")
    else:
        print(f"No section '{args.section_name}' found. Available: {state['config'].sections()}")


def send(args):
    """Send a file to the configured AMQP queue"""

    data = None
    try:
        input_file = open(args.filename, 'rb')
        data = input_file.read()
    except IOError:
        print("File not found or wrong path")
        exit(0)

    # configparser section is not a real dictionary :-( **state["amqpconfig"]
    channel, connection = amqp_connect(state["amqpconfig"]["username"],
                           state["amqpconfig"]["password"],
                           state["amqpconfig"]["host"],
                           state["amqpconfig"]["port"],
                           state["amqpconfig"]["ssl_enabled"])
    try:
        # enable delivery confirmations
        channel.confirm_delivery()
        # write to channel
        channel.basic_publish(
            exchange=state["amqpconfig"]["exchange"],
            routing_key=state["amqpconfig"]["routing_key"],
            body=data,
            mandatory=True,
            properties=pika.BasicProperties(
                delivery_mode=2,  # make message persistent
                user_id=state["amqpconfig"]["username"]
            ))
    except pika.exceptions.UnroutableError as e:
        exit(f"Failed to send data: {e}")
    except pika.exceptions.AMQPConnectionError as e:
        exit(f"Failed to send data, connection failed: {e}")
    except pika.exceptions.AMQPChannelError as e:
        exit(f"Failed to send data, channel is dead: ({e})")
    except Exception as e:
        exit(f"Failed to send data: ({e})")

    print(" [x] Sent %r" % os.path.basename(args.filename))
    connection.close()


def revoketoken(args):
    """Revoke the meteohub token locally stored"""
    mh = MhClient(state["mhconfig"], state["notoken"], state["verbose"])
    mh.connect()
    mh.revoketoken()
    mh.disconnect()


def reqls(args):
    """List all the meteohub request instances of the user"""
    mh = MhClient(state["mhconfig"], state["notoken"], state["verbose"])
    mh.connect()
    rqs = mh.datarequests()
    for rq in rqs:
        if rq.get("schedule_id", None) is None: # skip scheduled requests
            print(f"|{rq['name']}|{rq.get('fileoutput', '')}|")
    mh.disconnect()


def reqdl(args):
    """Download the output of a meteohub request instance"""

    if args.search_key is None or args.search_key == "":
        raise InvalidSearchKey()
    exitcode = 1
    mh = MhClient(state["mhconfig"], state["notoken"], state["verbose"])
    mh.connect()
    rqs = mh.datarequests()
    for rq in rqs:
        if rq["name"] == args.search_key and rq.get("schedule_id", None) is None:
            if state["verbose"]:
                print(f"found {args.search_key} with {rq.get('fileoutput', 'no file')} to download")
            if rq.get('fileoutput', None) is not None:
                dldata(mh, rq["fileoutput"], rq["name"]+"|")
            break
    else:
        mh.disconnect()
        raise ResourceNotFound(args.search_key)
    mh.disconnect()


def schedls(args):
    """List all the meteohub scheduled request instances of the user"""
    mh = MhClient(state["mhconfig"], state["notoken"], state["verbose"])
    mh.connect()
    rqs = mh.datarequests()
    for rq in rqs:
        if rq.get("schedule_id", None) is not None: # only scheduled requests
            rft = parsemhdate(rq["args"]["reftime"]["to"])
            print(f"|{rq['name']}|{rq.get('fileoutput', '')}|{formdate(rft)}|")
    mh.disconnect()
#    rqs = mh.schedules()


def scheddl(args):
    """Download the output of a meteohub scheduled request instance"""
    # considerare di usare /schedules/<schedule_id>/requests?last=True

    name, srft = parse_search_key(args.search_key)
    if state["verbose"]:
        print("Matching reference time "+str(srft))

    mh = MhClient(state["mhconfig"], state["notoken"], state["verbose"])
    mh.connect()
    rqs = mh.datarequests()
    for rq in rqs:
        if rq["name"] == name and rq.get("schedule_id", None) is not None:
            rft = parsemhdate(rq["args"]["reftime"]["to"])
            if matchdate(srft, rft):
                if state["verbose"]:
                    print(f"found {args.search_key} with {rq.get('fileoutput', 'no file')} to download")
                if rq.get('fileoutput', None) is not None:
                    dldata(mh, rq["fileoutput"], f"{rq['name']}|{formdate(rft)}|")
                break
    else:
        mh.disconnect()
        raise ResourceNotFound(args.search_key)
    mh.disconnect()


def schedrm(args):
    """Remove a scheduled request instance and its associated output data"""

    name, srft = parse_search_key(args.search_key)

    mh = MhClient(state["mhconfig"], state["notoken"], state["verbose"])
    mh.connect()
    rqs = mh.datarequests()
    rqs.reverse() # to remove the oldest
    for rq in rqs:
        if rq["name"] == name and rq.get("schedule_id", None) is not None:
            rft = parsemhdate(rq["args"]["reftime"]["to"])
            if matchdate(srft, rft):
                if state["verbose"]:
                    print(f"found {args.search_key} to delete")
                mh.datarequestdelete(str(rq["id"]))
                print(f"|{rq['name']}|{formdate(rft)}|")
                break
    mh.disconnect()


def data(args):
    """Download a meteohub resource by name"""
    mh = MhClient(state["mhconfig"], state["notoken"], state["verbose"])
    mh.connect()
    dldata(mh, args.name)
    mh.disconnect()


def dsls(args):
    """List all the datasets available on the server"""

    mh = MhClient(state["mhconfig"], state["notoken"], state["verbose"])
    mh.connect()
    rqs = mh.datasets()
    for rq in rqs:
        print(f"|{rq['name']}|{rq['category']}|{rq['format']}|")
# description, is_public
    mh.disconnect()


def odls(args):
    """List opendata files available for a specific dataset for all
    reference times"""

    mh = MhClient(state["mhconfig"], state["notoken"], state["verbose"])
    mh.connect()
    rqs = mh.odata(args.dataset)
    for rq in rqs:
        f = rq.get("filename", None)
        if f is not None:
            rft = parsemhdate(rq["date"]+"T"+rq.get("run", "00:00")+":00.0")
            print(f"|{formdate(rft)}|{f}|")
    mh.disconnect()


def odata(args):
    """Download a meteohub opendata resource by name"""

    mh = MhClient(state["mhconfig"], state["notoken"], state["verbose"])
    mh.connect()
    dldata(mh, args.name, opend=True)
    mh.disconnect()


def oddl(args):
    """Download a meteohub opendata resource for a specific dataset and reference time"""

    dataset, srft = parse_search_key(args.search_key)
    if state["verbose"]:
        print("Matching reference time "+str(srft))

    mh = MhClient(state["mhconfig"], state["notoken"], state["verbose"])
    mh.connect()
    rqs = mh.odata(dataset)
    for rq in rqs:
        f = rq.get("filename", None)
        if f is not None:
            rft = parsemhdate(rq["date"]+"T"+rq.get("run", "00:00")+":00.0")
            if matchdate(srft, rft):
                if state["verbose"]:
                    print(f"found {args.search_key} with {f} to download")
                dldata(mh, f, f"{dataset}|{formdate(rft)}|", opend=True)
                break


            mh.disconnect()


def dldata(mh, name, extralog="", opend=False):
    if name is None: raise InvalidResourceName()
    if mh.data(name, state.get("output", name), opend) > 0:
        print(f"|{name}|{state.get('output', name)}|{extralog}")


def dlcallback(ch, method, properties, body):
    if not state["quiet"]:
        print("Received message at %s" % datetime.datetime.now().isoformat())
    if state["verbose"]:
        print("Contents: %r" % body)
    try:
        msg = json.loads(body)
        if msg.get("filename", None) is not None:
            mh = MhClient(state["mhconfig"], state["notoken"], state["verbose"])
            mh.connect()
            dldata(mh, msg["filename"], f"{msg['request_name']}|{formdate(parsemhdate(msg['reftime']['to']))}|")
            mh.disconnect()
    except:
        if not state["quiet"]:
            print("Message does not look like a on-data-ready event")

def idlecallback(ch, method, properties, body):
    if not state["quiet"]:
        print("Received message at %s" % datetime.datetime.now().isoformat())
    if state["verbose"]:
        print("Contents: %r" % body)
    try:
        msg = json.loads(body)
        if msg.get("filename", None) is not None:
            print(f"|{msg['filename']}|")
    except:
        if not state["quiet"]:
            print("Message does not look like a on-data-ready event")


def waitandx(cb):
    channel, connection = amqp_connect(state["amqpconfig"]["username"],
                           state["amqpconfig"]["password"],
                           state["amqpconfig"]["host"],
                           state["amqpconfig"]["port"],
                           state["amqpconfig"]["ssl_enabled"])
    channel.basic_consume(queue=state["amqpconfig"]["queue"],
                          on_message_callback=cb,
                          auto_ack=True)
    print('Waiting for messages. To exit press CTRL+C')
    try:
        channel.start_consuming()
    except KeyboardInterrupt:
        print('Interrupted')
        try:
            sys.exit(0)
        except SystemExit:
            os._exit(0)
    connection.close()


def version(args):
    print(f"meteo-hub-cli version {VERSION}")
    sys.exit(0)


def waitanddl(args):
    """Wait for on-data-ready signal and download data"""
    waitandx(dlcallback)


def wait(args):
    """Wait for on-data-ready signal and print id of available data"""
    waitandx(idlecallback)


def main():
    parser = argparse.ArgumentParser(
        description="Command-line client for the Meteohub service")

    # add global command-line options
    parser.add_argument("--configfile", default="config.ini", metavar="FILE",
                        help="Use alternative configuration file (default: %(default)s)")
    parser.add_argument("--verbose", action="store_true", help="Increase verbosity")
    parser.add_argument("--quiet", action="store_true", help="Decrease verbosity")
    parser.add_argument("--notoken", action="store_true", help="Do not use saved token for logging in")
    parser.add_argument("--output", default=None, metavar="FILE",
                        help="Output file name for saving downloaded data (default: same name as on server)")

    # add subparser for the subcommands, consider to add keys:
    # description="", help="", required=True
    # the latter is useful but available only in python >= 3.7, not in epel8
    subparsers = parser.add_subparsers(title="list of subcommands", metavar=None)

    # add positional commands, their options, if any, and the action callback
    parser_settings = subparsers.add_parser("version", help="Print program version")
    parser_settings.set_defaults(func=version)

    parser_settings = subparsers.add_parser("settings", help="Print program configuration")
    parser_settings.add_argument("section_name", help="section to print")
    parser_settings.set_defaults(func=settings)

    parser_revoketoken = subparsers.add_parser("revoketoken", help="Revoke the meteohub authentication token locally stored")
    parser_revoketoken.set_defaults(func=revoketoken)

    parser_reqls = subparsers.add_parser("reqls", help="List all the meteohub request instances of the user")
    parser_reqls.set_defaults(func=reqls)

    parser_reqdl = subparsers.add_parser("reqdl", help="Download the output of a meteohub request instance")
    parser_reqdl.add_argument("search_key", help="name of the request to download")
    parser_reqdl.set_defaults(func=reqdl)

    parser_schedls = subparsers.add_parser("schedls", help="List all the meteohub scheduled request instances of the user")
    parser_schedls.set_defaults(func=schedls)

    parser_scheddl = subparsers.add_parser("scheddl", help="Download the output of a meteohub scheduled request instance")
    parser_scheddl.add_argument("search_key", help="request to download in the form name[:reftime], where reftime is YYYYmmdd[HHMM], if reftime is missing the last available is downloaded")
    parser_scheddl.set_defaults(func=scheddl)

    parser_schedrm = subparsers.add_parser("schedrm", help="Remove a scheduled request instance and its associated output data")
    parser_schedrm.add_argument("search_key", help="request to remove in the form name[:reftime], where reftime is YYYYmmdd[HHMM], if reftime is missing the oldest available is removed")
    parser_schedrm.set_defaults(func=schedrm)

    parser_data = subparsers.add_parser("data", help="Download a meteohub resource by name")
    parser_data.add_argument("name", help="name of the resource to be downloaded, without url part")
    parser_data.set_defaults(func=data)

    parser_dsls = subparsers.add_parser("dsls", help="List all the datasets available on the server")
    parser_dsls.set_defaults(func=dsls)

    parser_odls = subparsers.add_parser("odls", help="List opendata files available for a specific dataset for all reference times")
    parser_odls.add_argument("dataset", help="dataset for which a list of opendata files is requested")
    parser_odls.set_defaults(func=odls)

    parser_odata = subparsers.add_parser("odata", help="Download a meteohub opendata resource by name")
    parser_odata.add_argument("name", help="opendata resource to download, without url part")
    parser_odata.set_defaults(func=odata)

    parser_oddl = subparsers.add_parser("oddl", help="Download a meteohub opendata resource for a specific dataset and reference time")
    parser_oddl.add_argument("search_key", help="resource to download in the form dataset[:reftime], where reftime is YYYYmmdd[HHMM], if reftime is missing the last available is downloaded")
    parser_oddl.set_defaults(func=oddl)

    parser_waitanddl = subparsers.add_parser("waitanddl", help="Wait for on-data-ready signal from the configured AMQP queue and download data")
    parser_waitanddl.set_defaults(func=waitanddl)

    parser_wait = subparsers.add_parser("wait", help="Wait for on-data-ready signal from the configured AMQP queue and print id of available data")
    parser_wait.set_defaults(func=wait)

    parser_send = subparsers.add_parser("send", help="Send a file to the configured AMQP queue")
    parser_send.add_argument("filename", help="name of file to send")
    parser_send.set_defaults(func=send)

    # parse the arguments provided
    args = parser.parse_args()

    # set global state according to command-line options
    state["configfile"] = args.configfile
    state["verbose"] = args.verbose and not args.quiet
    state["quiet"] = args.quiet and not args.verbose
    state["notoken"] = args.notoken
    if args.output: state["output"] = args.output
    # read config, improve management of globals
    state["config"] = configparser.ConfigParser()
    state["config"].read(state["configfile"])
    state["amqpconfig"] = state["config"]["AMQP"]
    state["mhconfig"] = state["config"]["Meteohub"]
    # execute requested command
    args.func(args)

if __name__ == "__main__":
    main()
