import os
import stat
import json
import requests


class MhClient:
    class AuthorizationError(Exception):
        pass

    class RedirectionError(Exception):
        pass

    class SecretFilePermissionError(Exception):
        pass

    def __init__(self, config, notoken=False, verbose=False):
        self.config = config
        self.tokenfile = ".mhtoken"
        self.notoken = notoken
        self.verbose = verbose
        self.headers = None
        self.recycledtoken = False


    def __del__(self):
        self.headers = None
        self.recycledtoken = False


    def _makeheaders(self):
        """Internal method for creating the http headers.
        """
        self.headers = {"Authorization": "Bearer " + self.token,
                        "Content-Type": "text/plain"}


    def _processreq(self, r):
        if self.verbose:
            print("http status code: "+str(r.status_code))
        if r.status_code == 401: # unauthorized
            raise MhClient.AuthorizationError("Token invalid")
        elif  r.status_code == 301: # redirected
            raise MhClient.RedirectionError("Redirected, please update URL in config file")
        return r


    def _authget(self, api, **kwargs):
        """Internal method for performing a meteohub API get request. When an
        old token is used, in case of authentication failure a full
        login is tried once.
        
        :param api: the api url
        :param **kwargs: extra argument to requests.get method
        """
        r = requests.get(self.url+api, headers=self.headers,
                         allow_redirects=False, **kwargs)
        if r.status_code == 401: # unauthorized
            if self.recycledtoken: # try once a full login to get a new token
                if self.verbose:
                    print("Old token apparently expired, trying to get a new one")
                self.connect()
                r = requests.get(self.url+api, headers=self.headers,
                                 allow_redirects=False, **kwargs)
        return self._processreq(r)


    # merge with _authget to avoid code repetition
    def _authdelete(self, api, **kwargs):
        """Internal method for performing a meteohub API delete request. When
        an old token is used, in case of authentication failure a full
        login is tried once.
        
        :param api: the api url
        :param **kwargs: extra argument to requests.get method
        """
        r = requests.delete(self.url+api, headers=self.headers,
                            allow_redirects=False, **kwargs)
        if r.status_code == 401: # unauthorized
            if self.recycledtoken: # try once a full login to get a new token
                if self.verbose:
                    print("Old token apparently expired, trying to get a new one")
                self.connect()
                r = requests.delete(self.url+api, headers=self.headers,
                                 allow_redirects=False, **kwargs)
        return self._processreq(r)


    def check_configfile_mode(self, filecheck):
        """Check that the configuration file has strict permissions.
        """
        mode = stat.S_IMODE(os.stat(filecheck)[stat.ST_MODE])
        if mode & (stat.S_IRWXO | stat.S_IRWXG) != 0:
            raise MhClient.SecretFilePermissionError("File "+filecheck+" must not have rwx permissions for group and other")


    def retrievetoken(self):
        """Retrieve token for the current server from the configured
        tokenfile, if any
        """
        try:
            tkf = json.load(open(self.tokenfile))
            self.token = tkf.get(self.url, None)
        except:
            self.token = None


    def savetoken(self):
        """Save the current token in the tokenfile in json format, merging it
        with other tokens already present for different servers
        """
        try:
            tkf = json.load(open(self.tokenfile))
        except:
            tkf = {}
        if self.token is None:
            try:
                d.pop(self.url, None)
            except:
                pass
        else:
            tkf[self.url] = self.token
        json.dump(tkf, open(self.tokenfile, "w"), indent=0)


    def connect(self):
        """Connect to the system using the token from the configuration file
        if available (nothing is done in that case), or by logging in
        with the credentials from configuration file thus getting a
        new token.
        """
        self.url = self.config.get("url", "https://meteohub.mistralportal.it/")
        if self.verbose:
            print("Meteohub URL: "+self.url)
        if not self.notoken and not self.recycledtoken: # do not recycle twice
            self.retrievetoken()
            if self.token is not None:
                self.recycledtoken = True
                self._makeheaders()
                if self.verbose:
                    print("Recycling old token from "+self.tokenfile)
                return
        if self.verbose:
            print("Logging in as user: "+self.config["username"])
        r = requests.post(self.url+"auth/login",
                          {"username": self.config["username"],
                           "password": self.config["password"]},
                          allow_redirects=False)
        if r.status_code == 200:
            self.token = r.json()
            self._makeheaders()
            if not self.notoken: # update token file with new token
                self.savetoken()
            if self.verbose:
                print("Logged in succesfully")
        else:
            raise MhClient.AuthorizationError("Authorization failed for user "+self.config["username"])


    def revoketoken(self):
        """Revoke the current token by logging out and removing the
        corresponding entry from the configuration file.
        """
        self.logout()
        self.token = None
        self.savetoken()


    def disconnect(self):
        """Disconnect from the system without logging out unless explicitly
        requested with --notoken. This is the correct method to be called
        for final cleanup"""
        if self.headers:
            if self.notoken:
                self.logout()


    def logout(self):
        """Force an API logout with current token. This should not normally be
        called for cleanup.
        """
        if self.headers:
            r = requests.get(self.url+"auth/logout", headers=self.headers)
            self.headers = None
            if self.verbose:
                print("Logged out succesfully")


# Meteohub API methods
    def datarequests(self):
        r = self._authget("api/requests", params={"size": "100"})
        if r.status_code == 200: # or r.status_code == 206:
            return r.json()

    def schedules(self):
        r = self._authget("api/schedules", params={"size": "100"})
        if r.status_code == 200: # or r.status_code == 206:
            return r.json()

    def datarequestdelete(self, reqid):
        r = self._authdelete("api/requests/"+reqid)
        if r.status_code == 200:
            return r.json()
        if r.status_code == 404:
            if self.verbose:
                print("Request does not exist")
            return {}

    def scheduledelete(self, reqid):
        r = self._authdelete("api/schedules/"+reqid)
        if r.status_code == 200:
            return r.json()
        if r.status_code == 404:
            if self.verbose:
                print("Request does not exist")
            return {}

    def datasets(self):
        r = self._authget("api/datasets")
        if r.status_code == 200:
            return r.json()

    def data(self, filename, dest, opend):
        if opend: apif = "opendata/"
        else: apif = "data/"
        if self.verbose:
            print("Downloading "+filename+" to "+dest)
        r = self._authget("api/" + apif + filename, stream=True)
        if r.status_code == 200:
            filsiz = 0
            with open(dest, 'wb') as fd:
                for chunk in r.iter_content(chunk_size=1024*1024):
                    fd.write(chunk)
                    filsiz = filsiz + len(chunk)
                    if self.verbose:
                        print(".", end="", flush=True)
            if self.verbose:
                print()
                print("Download terminated")
            return filsiz
        else:
            if r.status_code == 404:
                if self.verbose:
                    print("File does not exist")
            else:
                if self.verbose:
                    print("Download failed")
            return -1

    def odata(self, ds):
        r = self._authget("api/datasets/" + ds + "/opendata")
        if r.status_code == 200:
            return r.json()
        if r.status_code == 404:
            if self.verbose:
                print("Dataset does not exist")
            return {}
